#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>

#include <linux/version.h>
#include <asm/uaccess.h>

#include "video_decoder.h"
#include "zr36120.h"
#include "common.h"

typedef unsigned char uchar;

static unsigned int remap=0;    /* remap Zoran */
       unsigned int vidmem=0;
       unsigned int triton1=0;

MODULE_PARM(remap,"i");
MODULE_PARM(vidmem,"i");
MODULE_PARM(triton1,"i");

/* Anybody who uses more than four? */
#define ZORAN_MAX 4

static int zoran_num;		    /* number of Zortecs in use */
static struct zoran zorans[ZORAN_MAX];

struct { int mode; int bpp; } palette2fmt[] = {
/* n/a    */	{ 0, 0 },
/* GREY   */	{ 0, 0 },
/* HI240  */	{ 0, 0 },
/* RGB565 */	{ ZORAN_VFEC_RGB_RGB565|ZORAN_VFEC_LE, 2 },
/* RGB24  */	{ ZORAN_VFEC_RGB_RGB888|ZORAN_VFEC_LE|ZORAN_VFEC_PACK24, 3 },
/* RGB32  */	{ ZORAN_VFEC_RGB_RGB888|ZORAN_VFEC_LE, 4 },
/* RGB555 */	{ ZORAN_VFEC_RGB_RGB555|ZORAN_VFEC_LE, 2 },
/* YUV422 */	{ ZORAN_VFEC_RGB_YUV422|ZORAN_VFEC_LE, 1 },
/* YUYV   */	{ ZORAN_VFEC_RGB_YUV422              , 1 },
/* UYVY   */	{ ZORAN_VFEC_RGB_YUV422|ZORAN_VFEC_LE, 1 },
/* YUV420 */	{ 0, 0 }};
#define NRPALETTES (sizeof(palette2fmt)/sizeof(palette2fmt[0]))

/* ----------------------------------------------------------------------- */
/* ZORAN functions							   */
/* ----------------------------------------------------------------------- */

void zoran_dump(struct zoran *ztv)
{
	char	str[1024];
	char	*p=str; /* shut up, gcc! */
	int	i;

	for (i=0; i<0x60; i+=4) {
		if ((i % 16) == 0) {
			if (i) DEBUG(printk(KERN_DEBUG "%s\n",str));
			p = str;
			p+= sprintf(str, "       %04x: ",i);
		}
		p += sprintf(p, "%08x ",zrread(i));
	}
}

static
void zoran_irq(int irq, void *dev_id, struct pt_regs * regs)
{
	u32 stat,estat;
	int count = 0;
	struct zoran *ztv = (struct zoran *)dev_id;

static ulong code = 0;
static ulong irq0 = 0;
static ulong irq1 = 0;
static ulong lastjif = 0;
if (time_before(lastjif,jiffies)) {
	lastjif = jiffies+10*HZ;
//	DEBUG(printk(KERN_DEBUG "zoran: irq_code=%lu, irq0=%lu, irq1=%lu\n", code,irq0,irq1));
}

	for (;;) {
		/* get/clear interrupt status bits */
		stat=zrread(ZORAN_ISR);
		estat=stat & zrread(ZORAN_ICR);
		if (!estat)
			return;
		zrwrite(estat,ZORAN_ISR);
		IDEBUG(printk(KERN_DEBUG "zoran: estat %08x\n",estat));
		IDEBUG(printk(KERN_DEBUG "zoran:  stat %08x\n",stat));

		if (estat & ZORAN_ISR_CODE)
		{
			IDEBUG(printk(KERN_DEBUG "zoran: CodReplIRQ\n"));
			code++;
		}
		if (estat & ZORAN_ISR_GIRQ0)
		{
			IDEBUG(printk(KERN_DEBUG "zoran: GIRQ0\n"));
			irq0++;
		}
		if (estat & ZORAN_ISR_GIRQ1)
		{
			IDEBUG(printk(KERN_DEBUG "zoran: GIRQ1\n"));
			irq1++;
		}

		count++;
		if (count > 10)
			printk(KERN_WARNING "zoran: irq loop %d (%x)\n", count, estat);
		if (count > 20)
		{
			zrwrite(0, ZORAN_ICR);
			printk(KERN_ERR "zoran: IRQ lockup, cleared int mask\n");
		}
	}
}

/*
 *      Scan for a Zoran chip, request the irq and map the io memory
 */
static int find_zoran(void)
{
	unsigned char command, latency;
	int result;
	struct zoran *ztv;
	struct pci_dev *dev;

	zoran_num=0;

	if (!pcibios_present())
	{
		DEBUG(printk(KERN_DEBUG "zoran: PCI-BIOS not present or not accessible!\n"));
		return 0;
	}

	for (dev = pci_devices; dev != NULL; dev = dev->next)
	{
		if (dev->vendor != PCI_VENDOR_ID_ZORAN)
			continue;
		if (dev->device != PCI_DEVICE_ID_ZORAN_36120)
			continue;

		/* Ok, ZR36120 found! */
		ztv=&zorans[zoran_num];
		ztv->dev=dev;
		ztv->zoran_mem=NULL;

		pci_read_config_word (ztv->dev, PCI_DEVICE_ID,      &ztv->id);

		/* pci_read_config_dword(ztv->dev, PCI_BASE_ADDRESS_0, &ztv->zoran_adr); */
		ztv->zoran_adr = ztv->dev->base_address[0];
		if (remap&&(!zoran_num))
		{
			remap<<=20;
			remap&=PCI_BASE_ADDRESS_MEM_MASK;
			printk(KERN_INFO "Remapping to : 0x%08x.\n", remap);
			remap|=ztv->zoran_adr&(~PCI_BASE_ADDRESS_MEM_MASK);
			pci_write_config_dword(ztv->dev, PCI_BASE_ADDRESS_0, remap);
			pci_read_config_dword(ztv->dev,  PCI_BASE_ADDRESS_0, &ztv->zoran_adr);
			ztv->dev->base_address[0] = ztv->zoran_adr;
		}

		ztv->zoran_adr&=PCI_BASE_ADDRESS_MEM_MASK;
		pci_read_config_byte(ztv->dev, PCI_CLASS_REVISION,
			     &ztv->revision);
		printk(KERN_INFO "zoran: Zoran %x (rev %d) ",
			ztv->id, ztv->revision);
		printk("bus: %d, devfn: %d, ",
			ztv->dev->bus->number, ztv->dev->devfn);
		printk("irq: %d, ",ztv->dev->irq);
		printk("memory: 0x%08x.\n", ztv->zoran_adr);

		ztv->zoran_mem=ioremap(ztv->zoran_adr, 0x1000);

		result = request_irq(ztv->dev->irq, zoran_irq,
			SA_SHIRQ|SA_INTERRUPT,"zoran",(void *)ztv);
		if (result==-EINVAL)
		{
			printk(KERN_ERR "zoran: Bad irq number or handler\n");
			return -EINVAL;
		}
		if (result==-EBUSY)
		{
			printk(KERN_ERR "zoran: IRQ %d busy, change your PnP config in BIOS\n",ztv->dev->irq);
			return result;
		}
		if (result < 0)
			return result;

		/* Enable bus-mastering */
		pci_read_config_byte(ztv->dev, PCI_COMMAND, &command);
		command|=PCI_COMMAND_MASTER|PCI_COMMAND_MEMORY;
		pci_write_config_byte(ztv->dev, PCI_COMMAND, command);
		pci_read_config_byte(ztv->dev, PCI_COMMAND, &command);
		if (!(command&PCI_COMMAND_MASTER))
		{
			printk(KERN_ERR "zoran: PCI bus-mastering could not be enabled\n");
			return -1;
		}
		pci_read_config_byte(ztv->dev, PCI_LATENCY_TIMER, &latency);
		if (!latency)
		{
			latency=32;
			pci_write_config_byte(ztv->dev, PCI_LATENCY_TIMER, latency);
		}
		DEBUG(printk(KERN_DEBUG "zoran: latency: %02x\n", latency));
		zoran_num++;
	}
	if(zoran_num)
		printk(KERN_INFO "zoran: %d Zoran card(s) found.\n", zoran_num);
	return zoran_num;
}

static
int zoran_muxsel(struct zoran* ztv, int channel)
{
	int	rv;
	int	norm;

	i2c_control_device(&(ztv->i2c), I2C_DRIVERID_DECODER, DECODER_SET_INPUT, &channel);
	rv = i2c_control_device(&(ztv->i2c), I2C_DRIVERID_DECODER, DECODER_GET_NORM, &norm);
	if (!rv)
		ztv->norm = norm;
DEBUG(printk(KERN_DEBUG "       detected norm=%d\n",norm));
	return rv;
}

static
void zoran_cap(struct zoran* ztv, int on)
{
	DEBUG(printk(KERN_DEBUG "       zoran_cap(%d)\n",on));

	switch (on) {
	 case 0:	ztv->running = 0;
			zrand(~ZORAN_VDC_VIDEN,ZORAN_VDC);
			break;

	 case 1:	ztv->running = 1;
			zror(ZORAN_VDC_VIDEN,ZORAN_VDC);
zoran_dump(ztv);
			break;
	 case 2:	
#if 0
ztv->running = 2;
			zrwrite(kvirt_to_bus((ulong)ztv->fbuffer),ZORAN_VTOP);
			zrwrite(kvirt_to_bus((ulong)ztv->fbuffer+1024),ZORAN_VBOT);
//			zrwrite((1024<<16)|ZORAN_VSTR_SNAPSHOT,ZORAN_VSTR);
			zror(ZORAN_VDC_VIDEN,ZORAN_VDC);
			zrand(~ZORAN_VDC_VIDEN,ZORAN_VDC);
			zror(ZORAN_VSTR_GRAB, ZORAN_VSTR);
#endif
			break;
	}
}

static ulong dmask[] = {
	0xFFFFFFFF, 0xFFFFFFFE, 0xFFFFFFFC, 0xFFFFFFF8,
	0xFFFFFFF0, 0xFFFFFFE0, 0xFFFFFFC0, 0xFFFFFF80,
	0xFFFFFF00, 0xFFFFFE00, 0xFFFFFC00, 0xFFFFF800,
	0xFFFFF000, 0xFFFFE000, 0xFFFFC000, 0xFFFF8000,
	0xFFFF0000, 0xFFFE0000, 0xFFFC0000, 0xFFF80000,
	0xFFF00000, 0xFFE00000, 0xFFC00000, 0xFF800000,
	0xFF000000, 0xFE000000, 0xFC000000, 0xF8000000,
	0xF0000000, 0xE0000000, 0xC0000000, 0x80000000
};

static
void zoran_built_overlay(struct zoran* ztv, int count, struct video_clip *vcp)
{
	ulong*	mtop;
	int	ystep = (ztv->vidXshift + ztv->vidWidth+31)/32;	// next DWORD
	int	mult = ztv->interlace;		// double height?
	int	i;

	DEBUG(printk(KERN_DEBUG "       overlay at %p, ystep=%d, clips=%d\n",ztv->overlay,ystep,count));
	if (ztv->overlay == 0) {
		zrand(~ZORAN_OCR_OVLEN, ZORAN_OCR);
		return;
	}

for (i=0; i<count; i++) {
	struct video_clip *vp = vcp+i;
	DEBUG(printk(KERN_DEBUG "       %d: clip(%d,%d,%d,%d)\n",
		i,vp->x,vp->y,vp->width,vp->height));
}

	/* clear entire blob */
	memset(ztv->overlay, 0, 1024*1024/8);

	/*
	 * activate the visible portion of the screen
	 * Note we take some shortcuts here, because we
	 * know the width can never be < 32. (I.e. a DWORD)
	 * We also assume the overlay starts somewhere in
	 * the FIRST dword.
	 */
	{
		int start = ztv->vidXshift;
		ulong firstd = dmask[start];
		ulong lastd = ~dmask[(start + ztv->width) & 31];
		mtop = ztv->overlay;
		for (i=0; i<ztv->height; i++) {
			int w = ztv->vidWidth;
			ulong* line = mtop;
			if (start & 31) {
				*line++ = firstd;
				w -= 32-(start&31);
			}
			memset(line, ~0, w/8);
			if (w & 31)
				line[w/32] = lastd;
			mtop += ystep;
		}
	}

	/* process clipping regions */
	for (i=0; i<count; i++) {
		int h;
		if (vcp->x<0 || vcp->x>ztv->width ||
		    vcp->y<0 || vcp->y>ztv->height ||
		    vcp->width <0 || (vcp->x+vcp->width) > ztv->width ||
		    vcp->height <0 || (vcp->y+vcp->height) > ztv->height)
		{
			DEBUG(printk(KERN_DEBUG "zoran: illegal clipzone (%d,%d,%d,%d)\n",vcp->x,vcp->y,vcp->width,vcp->height));
			continue;
		}

		mtop = &ztv->overlay[vcp->y*ystep];
		for (h=0; h<=vcp->height; h++) {
			int w;
			int x = ztv->vidXshift + vcp->x;
			for (w=0; w<=vcp->width; w++) {
				clear_bit(x&31, &mtop[x/32]);
				x++;
			}
			mtop += ystep;
		}
		++vcp;
	}

	mtop = ztv->overlay;
	zrwrite(virt_to_bus(mtop), ZORAN_MTOP);
	zrwrite(virt_to_bus(mtop+ystep), ZORAN_MBOT);
	zraor((mult*ystep)<<0,~ZORAN_OCR_MASKSTRIDE,ZORAN_OCR);
	zror(ZORAN_OCR_OVLEN, ZORAN_OCR);
}

static
void zoran_set_winsize(struct zoran* ztv)
{
	ulong	top, bot;
	int	stride;
	int	winWidth, winHeight;
	int	maxWidth, maxHeight, maxOffset;

	DEBUG(printk(KERN_DEBUG "       set_winsize(screen(%d,%d), bpp=%d, depth=%d, bpl=%d, rect(%d,%d,%d,%d) norm=%d\n",
		ztv->swidth,ztv->sheight,ztv->bpp,ztv->depth,ztv->bpl,
		ztv->x,ztv->y,ztv->width,ztv->height,ztv->norm));

	if (ztv->running) {
		DEBUG(printk(KERN_DEBUG "       ALERT set_winsize while ON\n"));
		return;
	}

	switch (ztv->norm) {
	 case 1: /* BW60 */
	 case 2: /* NTSC */
		maxWidth = 640;
		maxHeight = 480;
		maxOffset = 125;
		break;
	 case 0: /* BW50 */
	 case 3: /* PAL */
	 case 4: /* SECAM */
		maxWidth = 768;
		maxHeight = 576;
		maxOffset = 101;
		break;
	 default:
		return;
	}

	/*
	 * set top, bottom ptrs. Since these must be DWORD aligned,
	 * posible adjust the x and the width of the window.
	 */
	ztv->vidXshift = 0;
	winWidth = ztv->width;
	top = ztv->vidadr + ztv->x*ztv->bpp + ztv->y*ztv->bpl;
	if (top & 3) {
		ztv->vidXshift = (top & 3) / ztv->bpp;
		winWidth += ztv->vidXshift;
		top &= ~3;
	}
DEBUG(printk(KERN_DEBUG "        window start shifted %d pixels left\n",ztv->vidXshift));
	bot = top + ztv->bpl;
	if (top & 3)
		printk(KERN_DEBUG "zoran: ALERT! top not DWORD aligned (%08lx)\n",bot);
	if (bot & 3)
		printk(KERN_DEBUG "zoran: ALERT! bot not DWORD aligned (%08lx)\n",bot);
	zrwrite(top,ZORAN_VTOP);
	zrwrite(bot,ZORAN_VBOT);

	/*
	 * Now make sure the vidWidth is DWORD aligned too,
	 * thereby making sure the stride to the next line
	 * is DWORD align too.
	 */
	if ((winWidth*ztv->bpp) & 3)
		winWidth += 4 - ((winWidth * ztv->bpp) & 3) / ztv->bpp;

	/* determine the DispMode and stride */
	if (ztv->height <= maxHeight/2) {
		/* single frame suffices for this height */
		zror(ZORAN_VFEC_DISPMOD, ZORAN_VFEC);
		ztv->interlace = 0;
		winHeight = ztv->height;
		stride = ztv->bpl - (winWidth*ztv->bpp);
	}
	else {
		/* interleaving needed for this height */
		zrand(~ZORAN_VFEC_DISPMOD, ZORAN_VFEC);
		ztv->interlace = 1;
		winHeight = ztv->height/2;
		stride = ztv->bpl*2 - (winWidth*ztv->bpp);
	}
	if (stride & 3)
		printk(KERN_DEBUG "zoran: ALERT! stride not DWORD aligned (%d)\n",stride);

	zraor((winHeight<<12)|(winWidth<<0),~(ZORAN_VDC_VIDWINHT|ZORAN_VDC_VIDWINWID), ZORAN_VDC);
	zraor(stride<<16,~ZORAN_VSTR_DISPSTRIDE,ZORAN_VSTR);

	/* remember vidWidth, vidHeight for overlay calculations */
	ztv->vidWidth = winWidth;
	ztv->vidHeight = winHeight;
DEBUG(printk(KERN_DEBUG "       top=%08lx, bottom=%08lx, winWidth=%d, winHeight=%d, stride=%d\n",
		top,bot,winWidth,winHeight,stride));

	/* determine scales and crops */
	if (1) {
		int Wa = maxWidth;
		int X = (winWidth*64+Wa-1)/Wa;
		int We = winWidth*64/X;
		int HorDcm = 64-X;
		int hcrop1 = 2*(Wa-We)/4;
		int hcrop2 = Wa-We-hcrop1;
		int Hstart = maxOffset + hcrop1;
		int Hend = maxOffset + Wa-1-hcrop2;

DEBUG(printk(KERN_DEBUG "       X: scale=%d, start=%d, end=%d\n", HorDcm, Hstart, Hend));

		zraor((Hstart<<10)|(Hend<<0),~(ZORAN_VFEH_HSTART|ZORAN_VFEH_HEND),ZORAN_VFEH);
		zraor((HorDcm<<14),~ZORAN_VFEC_HORDCM, ZORAN_VFEC);
	}
	if (1) {
		int Ha = maxHeight/2;
		int Y = (winHeight*64+Ha-1)/Ha;
		int He = winHeight*64/Y;
		int VerDcm = 64-Y;
		int vcrop1 = 2*(Ha-He)/4;
		int vcrop2 = Ha-He-vcrop1;
		int Vstart = 16 + vcrop1;
		int Vend = 16 + Ha-1-vcrop2;

DEBUG(printk(KERN_DEBUG "       Y: scale=%d, start=%d, end=%d\n", VerDcm, Vstart, Vend));
		zraor((Vstart<<10)|(Vend<<0),~(ZORAN_VFEV_VSTART|ZORAN_VFEV_VEND),ZORAN_VFEV);
		zraor((VerDcm<<8),~ZORAN_VFEC_VERDCM, ZORAN_VFEC);
	}

	// format
	zraor(ztv->color_fmt, ~(ZORAN_VFEC_RGB|ZORAN_VFEC_LE|ZORAN_VFEC_PACK24), ZORAN_VFEC);
}

/*
 * Open a zoran card. Right now the flags are just a hack
 */
static int zoran_open(struct video_device *dev, int flags)
{
	struct zoran *ztv = (struct zoran*)dev;
	int	users, i;

	DEBUG(printk(KERN_DEBUG "zoran: open(dev,%d)\n",flags));

	switch (flags) {
	 case 0:
		/* already active? */
		if (ztv->user)
			return -EBUSY;
		ztv->user++;

		/* unmute audio */

		/* count the users, when we are the absolute first
		 * user, find a vga card we can overlay on..
		 * (This is a deferred call, since at the time
		 * of the zoran_init the video does not have to
		 * be initialized...)
		 */
		users = 0;
		for (i=0; i<zoran_num; i++)
			users += zorans[i].user;
		if (users == 1) {
			unsigned int vidadr;
			if (find_vga(&vidadr))
				for (i=0; i<zoran_num; ++i)
					zorans[i].vidadr = vidadr;
		}

/******************************************
 We really should be doing lazy allocing...
 ******************************************/
		/* allocate a frame buffer */
		if (!ztv->fbuffer)
			ztv->fbuffer = rvmalloc(2*ZORAN_MAX_FBUF);
		if (!ztv->fbuffer) {
			/* could not get a buffer, bail out */
			ztv->user--;
			return -ENOBUFS;
		}
		/* at this time we _always_ have a framebuffer */
		memset(ztv->fbuffer,1,2*ZORAN_MAX_FBUF);

		if (!ztv->overlay)
			ztv->overlay = (void*)kmalloc(1024*1024/8, GFP_KERNEL);
		if (!ztv->overlay) {
			/* could not get a buffer, bail out */
			ztv->user--;
			rvfree(ztv->fbuffer, 2*ZORAN_MAX_FBUF);
			return -ENOBUFS;
		}
		/* at this time we _always_ have a overlay */

		/* setup the tuner to the initial values */
		i2c_control_device(&ztv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_PICTURE, &ztv->picture);

		/* default on CVBS2 since my camera resides there */
		zoran_muxsel(ztv, 1);
		break;
	 case 1:
		break;
	}
	MOD_INC_USE_COUNT;
	return 0;
}

static
void zoran_close(struct video_device* dev)
{
	struct zoran *ztv = (struct zoran*)dev;

	DEBUG(printk(KERN_DEBUG "zoran: close(dev)\n"));

	/* we are no longer active, goodbye */
	ztv->user--;

	/* mute audio */
	/* stop grabbing */
	zoran_cap(ztv, 0);

	/* free the allocated framebuffer */
	if (ztv->fbuffer)
		rvfree( ztv->fbuffer, 2*ZORAN_MAX_FBUF );
	ztv->fbuffer = 0;
	if (ztv->overlay)
		kfree( ztv->overlay );
	ztv->overlay = 0;

	MOD_DEC_USE_COUNT;
}

static
long zoran_write(struct video_device* dev, const char* buf, unsigned long count, int nonblock)
{
        DEBUG(printk(KERN_DEBUG "zoran_write\n"));
	return -EINVAL;
}

static
long zoran_read(struct video_device* dev, char* buf, unsigned long count, int nonblock)
{
	struct zoran *ztv = (struct zoran*)dev;
	int	q,todo;

        DEBUG(printk(KERN_DEBUG "zoran_read(%p,%ld,%d)\n",buf,count,nonblock));
	todo = count;
	while (todo) {
		q = 1024;
		if (copy_to_user((void*)buf, (void*)ztv->fbuffer, q))
			return -EFAULT;
		todo -= q;
		buf += q;
	}

	return count;
}

static
void new_clip(struct video_window* vw, struct video_clip* vcp, int x, int y, int w, int h)
{
	vcp[vw->clipcount].x = x;
	vcp[vw->clipcount].y = y;
	vcp[vw->clipcount].width = w;
	vcp[vw->clipcount].height = h;
	vw->clipcount++;
}

static
int zoran_ioctl(struct video_device* dev, unsigned int cmd, void *arg)
{
	struct zoran* ztv = (struct zoran*)dev;

	switch (cmd) {
	 case VIDIOCGCAP:
	 {	/* get video capabilities */
		struct video_capability b;
		DEBUG(printk(KERN_DEBUG "zoran: Get Capabilities\n"));
		strcpy(b.name,ztv->video_dev.name);
		b.type = VID_TYPE_CAPTURE|
			 VID_TYPE_OVERLAY|
			 VID_TYPE_CLIPPING|
			 VID_TYPE_FRAMERAM|
			 VID_TYPE_SCALES;
		b.channels = 2;	/* cvbs, svhs */
		b.audios = 0;
		b.maxwidth = 768;
		b.maxheight = 576;
		b.minwidth = 32;
		b.minheight = 32;
		if (copy_to_user(arg,&b,sizeof(b)))
			return -EFAULT;
		return 0;
	 }

	 case VIDIOCGCHAN:
	 {
		struct video_channel v;
		DEBUG(printk(KERN_DEBUG "zoran: GetChannel\n"));
		if (copy_from_user(&v, arg,sizeof(v)))
			return -EFAULT;
		v.flags=VIDEO_VC_AUDIO;
		v.tuners=0;
		v.type=VIDEO_TYPE_CAMERA;
		switch (v.channel) {
		 case 0:
			strcpy(v.name,"CVBS");
			break;
		 case 1:
			strcpy(v.name,"SVHS");
			break;
		 default:
			return -EINVAL;
		}
		if (copy_to_user(arg,&v,sizeof(v)))
			return -EFAULT;
		return 0;
	 }
	 case VIDIOCSCHAN:
	 {	/* set video channel */
		int v;
		get_user_ret(v, (int*)arg, -EFAULT);
		DEBUG(printk(KERN_DEBUG "zoran: SetChannel(%d)\n",v));
		switch (v) {
		 case 0:
			v = 1;	/* the Composite input */
			break;
		 case 1:
			v = 6;	/* the SVHS input */
			break;
		 default:
			return -EINVAL;
		}
		return zoran_muxsel(ztv,v);
	 }

	 case VIDIOCGPICT:
	 {
		struct video_picture p = ztv->picture;
		DEBUG(printk(KERN_DEBUG "zoran: GetPicture\n"));
		p.depth = ztv->depth;
		switch (p.depth) {
		 case  8: p.palette=VIDEO_PALETTE_YUV422;
			  break;
		 case 15: p.palette=VIDEO_PALETTE_RGB555;
			  break;
		 case 16: p.palette=VIDEO_PALETTE_RGB565;
			  break;
		 case 24: p.palette=VIDEO_PALETTE_RGB24;
			  break;
		 case 32: p.palette=VIDEO_PALETTE_RGB32;
			  break;
		}
		if (copy_to_user(arg, &p, sizeof(p)))
			return -EFAULT;
		return 0;
	 }
	 case VIDIOCSPICT:
	 {
		struct video_picture p;
		DEBUG(printk(KERN_DEBUG "zoran: SetPicture\n"));
		if (copy_from_user(&p, arg,sizeof(p)))
			return -EFAULT;

		/* depth must match with framebuffer */
		if (p.depth != ztv->depth)
			return -EINVAL;

		/* check if palette matches this bpp */
		if (p.palette<1 || p.palette>NRPALETTES ||
		    palette2fmt[p.palette].bpp != ztv->bpp)
			return -EINVAL;
printk(KERN_DEBUG "      color_fmt set to %d\n",p.palette);
//		ztv->color_fmt = palette2fmt[p.palette].mode;

		i2c_control_device(&ztv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_PICTURE, &p);
		ztv->picture = p;
		return 0;
	 }

	 case VIDIOCGWIN:
	 {
		struct video_window vw;
		DEBUG(printk(KERN_DEBUG "zoran: GetWindow\n"));
		vw.x      = ztv->x;
		vw.y      = ztv->y;
		vw.width  = ztv->width;
		vw.height = ztv->height;
		vw.chromakey= 0;
		vw.flags  = 0;
		if (ztv->interlace)
			vw.flags|=VIDEO_WINDOW_INTERLACE;
		if (copy_to_user(arg,&vw,sizeof(vw)))
			return -EFAULT;
		return 0;
	 }
	 case VIDIOCSWIN:
	 {
		struct video_window vw;
		struct video_clip *vcp;
		int on;

		if (copy_from_user(&vw,arg,sizeof(vw)))
			return -EFAULT;

		DEBUG(printk(KERN_DEBUG "zoran: SetWindow(%d,%d,%d,%d,%x,%d)\n",
			vw.x,vw.y,vw.width,vw.height,vw.flags,vw.clipcount));

		if (vw.flags)
			return -EINVAL;

		ztv->x      = vw.x;
		ztv->y      = vw.y;
		ztv->width  = vw.width;
		ztv->height = vw.height;

		on = ztv->running;

		zoran_cap(ztv,0);	// off
		zoran_set_winsize(ztv);

		if (vw.clipcount>256)
			return -EDOM;   /* Too many! */

		/*
		 *      Do any clips.
		 */
		vcp = vmalloc(sizeof(struct video_clip)*(vw.clipcount+4));
		if (vcp==NULL)
			return -ENOMEM;
		if (vw.clipcount && copy_from_user(vcp,vw.clips,sizeof(struct video_clip)*vw.clipcount))
			return -EFAULT;

		/*
		 *      Impose display clips
		 */
		if (ztv->x<0)
			new_clip(&vw, vcp, 0, 0, -ztv->x, ztv->height-1);
		if (ztv->y<0)
			new_clip(&vw, vcp, 0, 0, ztv->width-1,-ztv->y);
		if (ztv->x+ztv->width> ztv->swidth)
			new_clip(&vw, vcp, ztv->swidth-ztv->x, 0, ztv->width-1, ztv->height-1);
		if (ztv->y+ztv->height > ztv->sheight)
			new_clip(&vw, vcp, 0, ztv->sheight-ztv->y, ztv->width-1, ztv->height-1);

		// built the requested clipping zones
		zoran_built_overlay(ztv, vw.clipcount, vcp);
		vfree(vcp);

		if (on && ztv->vidadr)		// on, when we were on
			zoran_cap(ztv,on);
		return 0;
	 }
	 case VIDIOCCAPTURE:
	 {
		int v;
		get_user_ret(v,(int*)arg, -EFAULT);
		DEBUG(printk(KERN_DEBUG "zoran: Capture(%d)\n",v));

		if (v==0)
			zoran_cap(ztv,0);	// off
		else {
			/* is VIDIOCSFBUF, VIDIOCSWIN done? */
			if (ztv->vidadr==0 || ztv->width==0 || ztv->height==0)
				return -EINVAL;

			zoran_cap(ztv,1);	// on
		}
		return 0;
	 }

	 case VIDIOCGFBUF:
	 {
		struct video_buffer v;
		DEBUG(printk(KERN_DEBUG "zoran: GetFramebuffer\n"));
		v.base   = (void *)ztv->vidadr;
		v.height = ztv->sheight;
		v.width  = ztv->swidth;
		v.depth  = ztv->depth;
		v.bytesperline = ztv->bpl;
		if(copy_to_user(arg, &v,sizeof(v)))
			return -EFAULT;
		return 0;
	 }
	 case VIDIOCSFBUF:
	 {
		struct video_buffer v;
		if (!capable(CAP_SYS_ADMIN))
			return -EPERM;
		if (copy_from_user(&v, arg,sizeof(v)))
			return -EFAULT;
		if (v.depth!=8 && v.depth!=15 && v.depth!=16 && v.depth!=24 && v.depth!=32)
			return -EINVAL;
		ztv->vidadr  = (unsigned long)v.base;
		ztv->sheight = v.height;
		ztv->swidth  = v.width;
		ztv->bpp     = ((v.depth+1)&0x38)/8;	/* bytes per pixel */
		ztv->depth   = v.depth;			/* bits per pixel */
		ztv->bpl     = v.bytesperline;

		DEBUG(printk(KERN_DEBUG "zoran: SetFrameBuffer(Display at %p is %d by %d, bytedepth %d, bpl %d)\n",
			v.base, v.width,v.height, ztv->bpp, ztv->bpl));
		zoran_set_winsize(ztv);
		return 0;
	 }

	 case VIDIOCMCAPTURE:
	 {
		struct video_mmap vm;
		DEBUG(printk(KERN_DEBUG "zoran: Mcapture\n"));
		if (copy_from_user(&vm,arg,sizeof(vm)))
			return -EFAULT;
		if (vm.frame<0 || vm.frame>1 ||
		    vm.width<0 || vm.width>768 ||
		    vm.height<0 || vm.height>576 ||
		    vm.format<0 || vm.format>3) //??
			return -EINVAL;

		zoran_cap(ztv,2);
		return 0;
	 }

	 case VIDIOCGTUNER:
	 case VIDIOCSTUNER:
	 case VIDIOCKEY:
	 case VIDIOCGFREQ:
	 case VIDIOCSFREQ:
	 case VIDIOCGAUDIO:
	 case VIDIOCSAUDIO:
		DEBUG(printk(KERN_DEBUG "zoran: unhandled vidioding(%x)\n",cmd));
		return 0;

	 default:
		DEBUG(printk(KERN_DEBUG "zoran: bad ioctl(%x)\n",cmd));
	}
	return -EPERM;
}

static
int zoran_initialise(struct video_device* dev)
{
        DEBUG(printk(KERN_DEBUG "zoran_initialise\n"));
	return 0;
}

static
int zoran_mmap(struct video_device* dev, const char* adr, unsigned long size)
{
	struct zoran* ztv = (struct zoran*)dev;
	unsigned long start = (unsigned long)adr;
	unsigned long pos;

        DEBUG(printk(KERN_DEBUG "zoran_mmap(%p,%ld)\n",adr,size));

	/* sanity checks */
	if (size > 2*ZORAN_MAX_FBUF || !ztv->fbuffer)
		return -EINVAL;

	/* start mapping the whole shabang to user memory */
	pos = (unsigned long)ztv->fbuffer;
	while (size>0) {
		unsigned long page = kvirt_to_phys(pos);
		if (remap_page_range(start, page, PAGE_SIZE, PAGE_READONLY))
			return -EAGAIN;
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	return 0;
}

static struct video_device zoran_template=
{
        "UNSET",
        VID_TYPE_TUNER|VID_TYPE_CAPTURE|VID_TYPE_OVERLAY,
        VID_HARDWARE_ZORAN,
        zoran_open,
        zoran_close,
        zoran_read,
        zoran_write,
        zoran_ioctl,
        zoran_mmap,
        zoran_initialise,
        NULL,
        0,
        0
};

static
int init_zoran(int i)
{
	struct zoran *ztv = &zorans[i];

	/* reset the zoran */
	zrand(~ZORAN_PCI_SOFTRESET,ZORAN_PCI);
	zror(ZORAN_PCI_SOFTRESET,ZORAN_PCI);

	DEBUG(printk(KERN_DEBUG "       zoran_mem: 0x%08x\n",(unsigned int) ztv->zoran_mem));

	/* default setup for max. PAL size in a 1024xXXX hicolor framebuffer */

	/* framegrabber details */
	ztv->vidadr=0;
	ztv->swidth=800;
	ztv->sheight=600;
	ztv->depth=16;
	ztv->bpp=2;
	ztv->bpl=ztv->swidth*ztv->bpp;

	/* more bla */
	ztv->x=0;
	ztv->y=0;
	ztv->width=768; /* 640 */
	ztv->height=576; /* 480 */
	ztv->norm=3;
	ztv->interlace=1;
	ztv->color_fmt=palette2fmt[VIDEO_PALETTE_RGB565].mode;

	/* maintenance data */
	ztv->fbuffer = NULL;
	ztv->overlay = NULL;
	ztv->user = 0;
	ztv->have_decoder = 0;
	ztv->running = 0;

	/* i2c */
	memcpy(&(ztv->i2c),&zoran_i2c_bus_template,sizeof(struct i2c_bus));
	sprintf(ztv->i2c.name,"zoran-%d",i);
	ztv->i2c.data = ztv;

	ztv->picture.colour=254<<7;
	ztv->picture.brightness=128<<8;
	ztv->picture.hue=128<<8;
	ztv->picture.contrast=0xd8<<7;

	zoran_set_winsize(ztv);		// setup sizes

	if (triton1)
		zror(ZORAN_VDC_TRICOM, ZORAN_VDC);

	/* external FL determines TOP frame */
	zror(ZORAN_VFEC_EXTFL, ZORAN_VFEC); 

	/* set HSpol */
	zror(ZORAN_VFEH_HSPOL, ZORAN_VFEH);
	/* set VSpol */
	zror(ZORAN_VFEV_VSPOL, ZORAN_VFEV);
	/* GenPur.8=output set to 1 */
	zraor((0x0<<0),~(0xFF<<0),ZORAN_PCI);
	zraor((0x00<<24),~(0xFF<<24),ZORAN_GUEST);
	
	/* clear interrupt status */
	zrwrite(0xFFFFFFFFUL, ZORAN_ISR);

	/*
	 * Now add the template and register the device unit
	 */
	memcpy(&ztv->video_dev, &zoran_template, sizeof(zoran_template));
	if (video_register_device(&ztv->video_dev, VFL_TYPE_GRABBER) < 0)
		return -1;
	i2c_register_bus(&ztv->i2c);

	/* set interrupt mask */
	zrwrite(ZORAN_ICR_GIRQ0|ZORAN_ICR_GIRQ1|ZORAN_ICR_CODE|ZORAN_ICR_EN, ZORAN_ICR);

	return 0;
}

static
void release_zoran(int max)
{
	u8 command;
	struct zoran *ztv;
	int i;

	for (i=0;i<max; i++) 
	{
		ztv=&zorans[i];

		/* turn off all capturing, DMA and IRQs */
		/* reset the zoran */
		zrand(~ZORAN_PCI_SOFTRESET,ZORAN_PCI);
		zror(ZORAN_PCI_SOFTRESET,ZORAN_PCI);

		/* first disable interrupts before unmapping the memory! */
		zrwrite(0, ZORAN_ICR);
		zrwrite(0xffffffffUL,ZORAN_ISR);

		/* free it */
                free_irq(ztv->dev->irq,ztv);
 
    		/* unregister i2c_bus */
		i2c_unregister_bus((&ztv->i2c));

		/* disable PCI bus-mastering */
		pci_read_config_byte(ztv->dev, PCI_COMMAND, &command);
 		command&=PCI_COMMAND_MASTER;
		pci_write_config_byte(ztv->dev, PCI_COMMAND, command);
    
		/* unmap and free memory */
		if (ztv->zoran_mem)
			iounmap(ztv->zoran_mem);

		video_unregister_device(&ztv->video_dev);
	}
}

#ifdef MODULE

int init_module(void)
{
#else
int init_zoran_cards(struct video_init *unused)
{
#endif
	int i;
  
	handle_chipset();
	if (find_zoran()<0)
		return -EIO;

	/* initialize Zorans */
	for (i=0; i<zoran_num; i++) {
		if (init_zoran(i)<0) {
			/* only release the zorans we have registered */
			release_zoran(i);
			return -EIO;
		} 
	}
	return 0;
}



#ifdef MODULE

void cleanup_module(void)
{
        release_zoran(zoran_num);
}
#endif
